default_index = {
    "settings": {
        "analysis": {
            'filter': {
                'text_ws_dsplit': {
                    'type': 'pattern_replace',
                    'pattern': r'(\.)',
                    'replacement': ' '
                }
            },
            "analyzer": {
                'string_ci': {
                    'type': 'custom',
                    'tokenizer': 'keyword',
                    'filter': ['lowercase']
                },
                "text_fuzzy": {
                    "type": "pattern",
                    "pattern": r"\s*:\s*",
                    "lowercase": False
                },
                "text_whitespace": {
                    "type": "whitespace"
                },
                'text_ws_dsplit': {
                    'type': 'custom',
                    'tokenizer': 'whitespace',
                    'filters': ['text_ws_dsplit']
                }
            },
            "normalizer": {
                "lowercase_normalizer": {
                    "type": "custom",
                    "char_filter": [],
                    "filter": ["lowercase"]
                }
            }
        }
    },
    'mappings': {}
}

default_mapping = {
    'dynamic': True,
    'properties': {
        '__text__': {'type': 'text'},
    }
}

default_dynamic_strings = {
    "strings_as_keywords": {
        "match_mapping_type": "string",
        "mapping": {
            "type": "keyword",
            "ignore_above": 8191
        }
    }
}

default_dynamic_templates = [
    {
        "int": {
            "path_match": "*_i",
            "mapping": {
                "type": "integer",
                "store": True
            }
        }
    },
    {
        "ints": {
            "path_match": "*_is",
            "mapping": {
                "type": "integer",
                "store": True
            }
        }
    },
    {
        "long": {
            "path_match": "*_l",
            "mapping": {
                "type": "long",
                "store": True
            }
        }
    },
    {
        "longs": {
            "path_match": "*_ls",
            "mapping": {
                "type": "long",
                "store": True
            }
        }
    },
    {
        "double": {
            "path_match": "*_d",
            "mapping": {
                "type": "float",
                "store": True
            }
        }
    },
    {
        "doubles": {
            "path_match": "*_ds",
            "mapping": {
                "type": "float",
                "store": True
            }
        }
    },
    {
        "float": {
            "path_match": "*_f",
            "mapping": {
                "type": "float",
                "store": True
            }
        }
    },
    {
        "floats": {
            "path_match": "*_fs",
            "mapping": {
                "type": "float",
                "store": True
            }
        }
    },
    {
        "string": {
            "path_match": "*_s",
            "mapping": {
                "type": "keyword",
                "store": True
            }
        }
    },
    {
        "strings": {
            "path_match": "*_ss",
            "mapping": {
                "type": "keyword",
                "store": True
            }
        }
    },
    {
        "text": {
            "path_match": "*_t",
            "mapping": {
                'type': 'text',
                "store": True
            }
        }
    },
    {
        "texts": {
            "path_match": "*_ts",
            "mapping": {
                'type': 'text',
                "store": True
            }
        }
    },
    {
        "boolean": {
            "path_match": "*_b",
            "mapping": {
                'type': 'boolean',
                "store": True
            }
        }
    },
    {
        "booleans": {
            "path_match": "*_bs",
            "mapping": {
                'type': 'boolean',
                "store": True
            }
        }
    },
    {
        "date": {
            "path_match": "*_dt",
            "mapping": {
                'type': 'date',
                'format': 'date_optional_time||epoch_millis',
                "store": True
            }
        }
    },
    {
        "date": {
            "path_match": "*_dts",
            "mapping": {
                'type': 'date',
                'format': 'date_optional_time||epoch_millis',
                "store": True
            }
        }
    }
]



from assemblyline import odm
from assemblyline.odm.messages import PerformanceTimer

MSG_TYPES = {"DispatcherHeartbeat"}
LOADER_CLASS = "assemblyline.odm.messages.dispatcher_heartbeat.DispatcherMessage"


@odm.model(description="Queue Model")
class Queues(odm.Model):
    ingest = odm.Integer(description="Number of submissions in ingest queue")
    start = odm.List(odm.Integer(), description="Number of submissions that started")
    result = odm.List(odm.Integer(), description="Number of results in queue")
    command = odm.List(odm.Integer(), description="Number of commands in queue")


@odm.model(description="Inflight Model")
class Inflight(odm.Model):
    max = odm.Integer(description="Maximum number of submissions")
    outstanding = odm.Integer(description="Number of outstanding submissions")
    per_instance = odm.List(odm.Integer(), description="Number of submissions per Dispatcher instance")


@odm.model(description="Metrics Model")
class Metrics(odm.Model):
    files_completed = odm.Integer(description="Number of files completed")
    submissions_completed = odm.Integer(description="Number of submissions completed")
    service_timeouts = odm.Integer(description="Number of service timeouts")
    cpu_seconds = PerformanceTimer(description="CPU time")
    cpu_seconds_count = odm.Integer(description="CPU count")
    busy_seconds = PerformanceTimer(description="Busy CPU time")
    busy_seconds_count = odm.Integer(description="Busy CPU count")
    save_queue = odm.Integer(description="Processed submissions waiting to be saved")
    error_queue = odm.Integer(description="Errors waiting to be saved")


@odm.model(description="Heartbeat Model")
class Heartbeat(odm.Model):
    inflight = odm.Compound(Inflight, description="Inflight submissions")
    instances = odm.Integer(description="Number of instances")
    metrics = odm.Compound(Metrics, description="Dispatcher metrics")
    queues = odm.Compound(Queues, description="Dispatcher queues")
    component = odm.Keyword(description="Component name")


@odm.model(description="Model of Dispatcher Heartbeat Messages")
class DispatcherMessage(odm.Model):
    msg = odm.Compound(Heartbeat, description="Heartbeat message")
    msg_loader = odm.Enum(values={LOADER_CLASS}, default=LOADER_CLASS, description="Loader class for message")
    msg_type = odm.Enum(values=MSG_TYPES, default="DispatcherHeartbeat", description="Type of message")
    sender = odm.Keyword(description="Sender of message")



import json
from copy import deepcopy


class ElasticBulkPlan(object):
    def __init__(self, indexes, model=None):
        self.indexes = indexes
        self.model = model
        self.operations = []

    @property
    def empty(self):
        return len(self.operations) == 0

    def add_delete_operation(self, doc_id, index=None):
        if index:
            self.operations.append(json.dumps({"delete": {"_index": index, "_id": doc_id}}))
        else:
            for cur_index in self.indexes:
                self.operations.append(json.dumps({"delete": {"_index": cur_index, "_id": doc_id}}))

    def add_insert_operation(self, doc_id, doc, index=None):
        if self.model and isinstance(doc, self.model):
            saved_doc = doc.as_primitives(hidden_fields=True)
        elif self.model:
            saved_doc = self.model(doc).as_primitives(hidden_fields=True)
        else:
            if not isinstance(doc, dict):
                saved_doc = {'__non_doc_raw__': doc}
            else:
                saved_doc = deepcopy(doc)
        saved_doc['id'] = doc_id

        self.operations.append(json.dumps({"create": {"_index": index or self.indexes[0], "_id": doc_id}}))
        self.operations.append(json.dumps(saved_doc))

    def add_upsert_operation(self, doc_id, doc, index=None):
        if isinstance(doc, self.model):
            saved_doc = doc.as_primitives(hidden_fields=True)
        elif self.model:
            saved_doc = self.model(doc).as_primitives(hidden_fields=True)
        else:
            if not isinstance(doc, dict):
                saved_doc = {'__non_doc_raw__': doc}
            else:
                saved_doc = deepcopy(doc)
        saved_doc['id'] = doc_id

        self.operations.append(json.dumps({"update": {"_index": index or self.indexes[0], "_id": doc_id}}))
        self.operations.append(json.dumps({"doc": saved_doc, "doc_as_upsert": True}))

    def add_update_operation(self, doc_id, doc, index=None):

        if isinstance(doc, self.model):
            saved_doc = doc.as_primitives(hidden_fields=True)
        elif self.model:
            saved_doc = self.model(doc, mask=list(doc.keys())).as_primitives(hidden_fields=True)
        else:
            if not isinstance(doc, dict):
                saved_doc = {'__non_doc_raw__': doc}
            else:
                saved_doc = deepcopy(doc)

        if index:
            self.operations.append(json.dumps({"update": {"_index": index, "_id": doc_id}}))
            self.operations.append(json.dumps({"doc": saved_doc}))
        else:
            for cur_index in self.indexes:
                self.operations.append(json.dumps({"update": {"_index": cur_index, "_id": doc_id}}))
                self.operations.append(json.dumps({"doc": saved_doc}))

    def get_plan_data(self):
        return "\n".join(self.operations)



from assemblyline import odm
from assemblyline.common import forge
from assemblyline.odm.models.statistics import Statistics

Classification = forge.get_classification()


@odm.model(index=True, store=True, description="Model of Service Heuristics")
class Heuristic(odm.Model):
    attack_id = odm.List(odm.Keyword(copyto="__text__"), default=[], description="List of all associated ATT&CK IDs")
    classification = odm.Classification(default=Classification.UNRESTRICTED,
                                        description="Classification of the heuristic")
    description = odm.Text(copyto="__text__", description="Description of the heuristic")
    filetype = odm.Keyword(copyto="__text__", description="What type of files does this heuristic target?")
    heur_id = odm.Keyword(copyto="__text__", description="ID of the Heuristic")
    name = odm.Keyword(copyto="__text__", description="Name of the heuristic")
    score = odm.Integer(description="Default score of the heuristic")
    signature_score_map = odm.Mapping(odm.Integer(), default={},
                                      description="Score of signatures for this heuristic")
    stats = odm.Compound(Statistics, default={}, description="Statistics related to the Heuristic")
    max_score = odm.Optional(odm.Integer(), description="Maximum score for heuristic")



import json
import time

from assemblyline.remote.datatypes import get_client, retry_call

_drop_card_script = """
local set_name = ARGV[1]
local key = ARGV[2]

redis.call('srem', set_name, key)
return redis.call('scard', set_name)
"""

_limited_add = """
local set_name = KEYS[1]
local key = ARGV[1]
local limit = tonumber(ARGV[2])

if redis.call('scard', set_name) < limit then
    redis.call('sadd', set_name, key)
    return true
end
return false
"""


class Set(object):
    def __init__(self, name, host=None, port=None):
        self.c = get_client(host, port, False)
        self.name = name
        self._drop_card = self.c.register_script(_drop_card_script)
        self._limited_add = self.c.register_script(_limited_add)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.delete()

    def add(self, *values):
        return retry_call(self.c.sadd, self.name,
                          *[json.dumps(v) for v in values])

    def limited_add(self, value, size_limit):
        """Add a single value to the set, but only if that wouldn't make the set grow past a given size."""
        return retry_call(self._limited_add, keys=[self.name], args=[json.dumps(value), size_limit])

    def exist(self, value):
        return retry_call(self.c.sismember, self.name, json.dumps(value))

    def length(self):
        return retry_call(self.c.scard, self.name)

    def members(self):
        return [json.loads(s) for s in retry_call(self.c.smembers, self.name)]

    def remove(self, *values):
        return retry_call(self.c.srem, self.name,
                          *[json.dumps(v) for v in values])

    def drop(self, value):
        return retry_call(self._drop_card, args=[value])

    def random(self, num=None):
        ret_val = retry_call(self.c.srandmember, self.name, num)
        if isinstance(ret_val, list):
            return [json.loads(s) for s in ret_val]
        else:
            return json.loads(ret_val)

    def pop(self):
        data = retry_call(self.c.spop, self.name)
        return json.loads(data) if data else None

    def pop_all(self):
        return [json.loads(s) for s in retry_call(self.c.spop, self.name, self.length())]

    def delete(self):
        retry_call(self.c.delete, self.name)


class ExpiringSet(Set):
    def __init__(self, name, ttl=86400, host=None, port=None):
        super(ExpiringSet, self).__init__(name, host, port)
        self.ttl = ttl
        self.last_expire_time = 0

    def _conditional_expire(self):
        if self.ttl:
            ctime = time.time()
            if ctime > self.last_expire_time + (self.ttl / 2):
                retry_call(self.c.expire, self.name, self.ttl)
                self.last_expire_time = ctime

    def add(self, *values):
        rval = super(ExpiringSet, self).add(*values)
        self._conditional_expire()
        return rval



from __future__ import annotations
import typing

from assemblyline.odm.base import *
from assemblyline.odm.base import Keyword, Optional, Boolean, Integer, List, Compound

_InnerType = typing.TypeVar("_InnerType")

"""
Helper functions to invoke ODM types without requiring type annotations.
"""


def description(text):
    def _fn(obj):
        obj.description = text
    return _fn


def keyword(*args, **kwargs) -> str:
    return typing.cast(str, Keyword(*args, **kwargs))


def optional(child_type: _InnerType, **kwargs) -> typing.Optional[_InnerType]:
    return typing.cast(typing.Optional[_InnerType], Optional(child_type, **kwargs))


def boolean(*args, **kwargs) -> bool:
    return typing.cast(bool, Boolean(*args, **kwargs))


def integer(*args, **kwargs) -> int:
    return typing.cast(int, Integer(*args, **kwargs))


def sequence(child_type: _InnerType, **kwargs) -> list[_InnerType]:
    return typing.cast(list[_InnerType], List(child_type, **kwargs))


def compound(child_type: typing.Callable[..., _InnerType], **kwargs) -> _InnerType:
    return typing.cast(_InnerType, Compound(child_type, **kwargs))





from assemblyline import odm
from assemblyline.odm.messages import PerformanceTimer

MSG_TYPES = {"IngestHeartbeat"}
LOADER_CLASS = "assemblyline.odm.messages.ingest_heartbeat.IngestMessage"


@odm.model(description="Queues")
class Queues(odm.Model):
    critical = odm.Integer(description="Size of the critical priority queue")
    high = odm.Integer(description="Size of the high priority queue")
    ingest = odm.Integer(description="Size of the ingest queue")
    complete = odm.Integer(description="Size of the complete queue")
    low = odm.Integer(description="Size of the low priority queue")
    medium = odm.Integer(description="Size of the medium priority queue")


@odm.model(description="Metrics")
class Metrics(odm.Model):
    cache_miss = odm.Integer(description="Number of cache misses")
    cache_expired = odm.Integer(description="Number of cache expires")
    cache_stale = odm.Integer(description="Number of cache stales")
    cache_hit_local = odm.Integer(description="Number of cache local hits")
    cache_hit = odm.Integer(description="Number of cache hits")
    bytes_completed = odm.Integer(description="Number of bytes completed")
    bytes_ingested = odm.Integer(description="Number of bytes ingested")
    duplicates = odm.Integer(description="Number of duplicate submissions")
    error = odm.Integer(description="Number of errors")
    files_completed = odm.Integer(description="Number of completed files")
    skipped = odm.Integer(description="Number of skipped files")
    submissions_completed = odm.Integer(description="Number of completed submissions")
    submissions_ingested = odm.Integer(description="Number of ingested submissions")
    timed_out = odm.Integer(description="Number of timed_out submissions")
    whitelisted = odm.Integer(description="Number of safelisted submissions")
    cpu_seconds = PerformanceTimer()
    cpu_seconds_count = odm.Integer()
    busy_seconds = PerformanceTimer()
    busy_seconds_count = odm.Integer()


@odm.model(description="Processing")
class Processing(odm.Model):
    inflight = odm.Integer(description="Number of inflight submissions")


@odm.model(description="Chance of Processing")
class ProcessingChance(odm.Model):
    critical = odm.Integer(description="Chance of processing critical items")
    high = odm.Integer(description="Chance of processing high items")
    low = odm.Integer(description="Chance of processing low items")
    medium = odm.Integer(description="Chance of processing medium items")


@odm.model(description="Heartbeat Model")
class Heartbeat(odm.Model):
    instances = odm.Integer(description="Number of ingest processes")
    metrics = odm.Compound(Metrics, description="Metrics")
    processing = odm.Compound(Processing, description="Inflight queue sizes")
    processing_chance = odm.Compound(ProcessingChance, description="Chance of processing items")
    queues = odm.Compound(Queues, description="Queue lengths block")


@odm.model(description="Model of Ingester Heartbeat Message")
class IngestMessage(odm.Model):
    msg = odm.Compound(Heartbeat, description="Heartbeat message")
    msg_loader = odm.Enum(values={LOADER_CLASS}, default=LOADER_CLASS, description="Loader class for message")
    msg_type = odm.Enum(values=MSG_TYPES, default="IngestHeartbeat", description="Type of message")
    sender = odm.Keyword(description="Sender of message")



import io

from math import ceil, log
from typing import Tuple, List, BinaryIO, AnyStr

frequency = None


def calculate_entropy(contents: bytes) -> float:
    """ this function calculates the entropy of the file
        It is given by the formula:
            E = -SUM[v in 0..255](p(v) * ln(p(v)))
    """
    calculator = BufferedCalculator()
    calculator.update(contents)
    return calculator.entropy()


def calculate_partition_entropy(fin: BinaryIO, num_partitions: int = 50) -> Tuple[float, List[float]]:
    """Calculate the entropy of a file and its partitions."""

    # Split input into num_parititions and calculate
    # parition entropy.
    fin.seek(0, io.SEEK_END)
    size = fin.tell()
    fin.seek(0)
    partition_size = int(ceil(size / float(num_partitions)))

    # Also calculate full file entropy using buffered calculator.
    p_entropies = []
    full_entropy_calculator = BufferedCalculator()
    for _ in range(num_partitions):
        partition = fin.read(partition_size)
        p_entropies.append(calculate_entropy(partition))
        full_entropy_calculator.update(partition)
    return full_entropy_calculator.entropy(), p_entropies


class BufferedCalculator(object):
    def __init__(self):
        global frequency
        import pyximport
        pyximport.install()
        # noinspection PyUnresolvedReferences
        from assemblyline.common import frequency

        self.c = {}
        self.length = 0

    def entropy(self) -> float:
        if self.length == 0:
            return 0.0

        length = float(self.length)

        entropy = 0.0
        for v in self.c.values():
            prob = float(v) / length
            entropy += prob * log(prob, 2)

        entropy *= -1

        # Make sure we don't return -0.0.
        if not entropy:
            entropy = 0.0

        return entropy

    def update(self, data: AnyStr, length: int = 0):
        if not length:
            length = len(data)

        self.length += length
        self.c = frequency.counts(data, length, self.c)



from __future__ import annotations
import json
import os
import random
import time
import threading
import logging
from typing import Any, Optional

from multiprocessing import Process

from assemblyline.common import forge
from assemblyline.common.uid import get_random_id
from assemblyline.odm.models.error import ERROR_TYPES
from assemblyline.remote.datatypes.hash import Hash
from assemblyline.remote.datatypes.queues.named import NamedQueue


# noinspection PyBroadException
def backup_worker(worker_id: str, instance_id: str, working_dir: str):
    datastore = forge.get_datastore(archive_access=True)
    worker_queue: NamedQueue[dict[str, Any]] = NamedQueue(f"r-worker-{instance_id}", ttl=1800)
    done_queue: NamedQueue[dict[str, Any]] = NamedQueue(f"r-done-{instance_id}", ttl=1800)
    hash_queue: Hash[str] = Hash(f"r-hash-{instance_id}")
    stopping = False
    with open(os.path.join(working_dir, "backup.part%s" % worker_id), "w+") as backup_file:
        while True:
            data = worker_queue.pop(timeout=1)
            if data is None:
                if stopping:
                    break
                continue

            if data.get('stop', False):
                if not stopping:
                    stopping = True
                else:
                    time.sleep(round(random.uniform(0.050, 0.250), 3))
                    worker_queue.push(data)
                continue

            missing = False
            success = True
            try:
                to_write = datastore.get_collection(data['bucket_name']).get(data['key'], as_obj=False)
                if to_write:
                    if data.get('follow_keys', False):
                        for bucket, bucket_key, getter in FOLLOW_KEYS.get(data['bucket_name'], []):
                            for key in getter(to_write.get(bucket_key, None)):
                                hash_key = "%s_%s" % (bucket, key)
                                if not hash_queue.exists(hash_key):
                                    hash_queue.add(hash_key, "True")
                                    worker_queue.push({"bucket_name": bucket, "key": key, "follow_keys": True})

                    backup_file.write(json.dumps((data['bucket_name'], data['key'], to_write)) + "\n")
                else:
                    missing = True
            except Exception:
                success = False

            done_queue.push({
                "success": success,
                "missing": missing,
                "bucket_name": data['bucket_name'],
                "key": data['key']
            })

    done_queue.push({"stopped": True})


# noinspection PyBroadException
def restore_worker(worker_id: str, instance_id: str, working_dir: str):
    datastore = forge.get_datastore(archive_access=True)
    done_queue: NamedQueue[dict[str, Any]] = NamedQueue(f"r-done-{instance_id}", ttl=1800)

    with open(os.path.join(working_dir, "backup.part%s" % worker_id), "rb") as input_file:
        for line in input_file:
            bucket_name, key, data = json.loads(line)

            success = True
            try:
                collection = datastore.get_collection(bucket_name)
                collection.save(key, data)
            except Exception:
                success = False

            done_queue.push({
                "success": success,
                "missing": False,
                "bucket_name": bucket_name,
                "key": key})

    done_queue.push({"stopped": True})


class DistributedBackup(object):
    def __init__(self, working_dir: str, worker_count: int = 50, spawn_workers: bool = True,
                 use_threading: bool = False, logger: Optional[logging.Logger] = None):
        self.working_dir = working_dir
        self.datastore = forge.get_datastore(archive_access=True)
        self.logger = logger
        self.plist: list[Process] = []
        self.use_threading = use_threading
        self.instance_id = get_random_id()
        self.worker_queue: NamedQueue[dict[str, Any]] = NamedQueue(f"r-worker-{self.instance_id}", ttl=1800)
        self.done_queue: NamedQueue[dict[str, Any]] = NamedQueue(f"r-done-{self.instance_id}", ttl=1800)
        self.hash_queue: Hash[str] = Hash(f"r-hash-{self.instance_id}")
        self.bucket_error: list[str] = []
        self.valid_buckets: list[str] = sorted(list(self.datastore.ds.get_models().keys()))
        self.worker_count = worker_count
        self.spawn_workers = spawn_workers
        self.total_count = 0
        self.error_map_count: dict[str, int] = {}
        self.missing_map_count: dict[str, int] = {}
        self.map_count: dict[str, int] = {}
        self.last_time: float = 0
        self.last_count = 0
        self.error_count = 0

    def cleanup(self):
        self.worker_queue.delete()
        self.done_queue.delete()
        self.hash_queue.delete()
        for p in self.plist:
            p.terminate()

    def done_thread(self, title: str):
        t0 = time.time()
        self.last_time = t0

        running_threads = self.worker_count

        while running_threads > 0:
            msg = self.done_queue.pop(timeout=1)

            if msg is None:
                continue

            if "stopped" in msg:
                running_threads -= 1
                continue

            bucket_name = msg.get('bucket_name', 'unknown')

            if msg.get('success', False):
                self.total_count += 1

                if msg.get("missing", False):
                    if bucket_name not in self.missing_map_count:
                        self.missing_map_count[bucket_name] = 0

                    self.missing_map_count[bucket_name] += 1
                else:
                    if bucket_name not in self.map_count:
                        self.map_count[bucket_name] = 0

                    self.map_count[bucket_name] += 1

                new_t = time.time()
                if (new_t - self.last_time) > 5:
                    if self.logger:
                        self.logger.info("%s (%s at %s keys/sec) ==> %s" %
                                         (self.total_count,
                                          new_t - self.last_time,
                                          int((self.total_count - self.last_count) / (new_t - self.last_time)),
                                          self.map_count))
                    self.last_count = self.total_count
                    self.last_time = new_t
            else:
                self.error_count += 1

                if bucket_name not in self.error_map_count:
                    self.error_map_count[bucket_name] = 0

                self.error_map_count[bucket_name] += 1

        # Cleanup
        self.cleanup()

        summary = ""
        summary += "\n########################\n"
        summary += "####### SUMMARY  #######\n"
        summary += "########################\n"
        summary += "%s items - %s errors - %s secs\n\n" % \
                   (self.total_count, self.error_count, time.time() - t0)

        for k, v in self.map_count.items():
            summary += "\t%15s: %s\n" % (k.upper(), v)

        if len(self.missing_map_count.keys()) > 0:
            summary += "\n\nMissing data:\n\n"
            for k, v in self.missing_map_count.items():
                summary += "\t%15s: %s\n" % (k.upper(), v)

        if len(self.error_map_count.keys()) > 0:
            summary += "\n\nErrors:\n\n"
            for k, v in self.error_map_count.items():
                summary += "\t%15s: %s\n" % (k.upper(), v)

        if len(self.bucket_error) > 0:
            summary += f"\nThese buckets failed to {title.lower()} completely: {self.bucket_error}\n"
        if self.logger:
            self.logger.info(summary)

    # noinspection PyBroadException,PyProtectedMember
    def backup(self, bucket_list: list[str], follow_keys: bool = False, query: Optional[str] = None):
        if query is None:
            query = 'id:*'

        for bucket in bucket_list:
            if bucket not in self.valid_buckets:
                if self.logger:
                    self.logger.warn("\n%s is not a valid bucket.\n\n"
                                     "The list of valid buckets is the following:\n\n\t%s\n" %
                                     (bucket.upper(), "\n\t".join(self.valid_buckets)))
                return

        targets = ', '.join(bucket_list)
        try:
            if self.logger:
                self.logger.info("\n-----------------------")
                self.logger.info("----- Data Backup -----")
                self.logger.info("-----------------------")
                self.logger.info(f"    Deep: {follow_keys}")
                self.logger.info(f"    Buckets: {targets}")
                self.logger.info(f"    Workers: {self.worker_count}")
                self.logger.info(f"    Target directory: {self.working_dir}")
                self.logger.info(f"    Filtering query: {query}")

            # Start the workers
            for x in range(self.worker_count):
                if self.use_threading:
                    t = threading.Thread(target=backup_worker, args=(x, self.instance_id, self.working_dir))
                    t.setDaemon(True)
                    t.start()
                else:
                    p = Process(target=backup_worker, args=(x, self.instance_id, self.working_dir))
                    p.start()
                    self.plist.append(p)

            # Start done thread
            dt = threading.Thread(target=self.done_thread, args=('Backup',), name="Done thread")
            dt.setDaemon(True)
            dt.start()

            # Process data buckets
            for bucket_name in bucket_list:
                try:
                    collection = self.datastore.get_collection(bucket_name)
                    for item in collection.stream_search(query, fl="id", item_buffer_size=500, as_obj=False):
                        self.worker_queue.push({"bucket_name": bucket_name, "key": item['id'],
                                                "follow_keys": follow_keys})

                except Exception as e:
                    self.cleanup()
                    if self.logger:
                        self.logger.exception(e)
                        self.logger.error("Error occurred while processing bucket %s." % bucket_name)
                    self.bucket_error.append(bucket_name)

            for _ in range(self.worker_count):
                self.worker_queue.push({"stop": True})

            dt.join()
        except Exception as e:
            if self.logger:
                self.logger.exception(e)

    def restore(self):
        try:
            if self.logger:
                self.logger.info("\n------------------------")
                self.logger.info("----- Data Restore -----")
                self.logger.info("------------------------")
                self.logger.info(f"    Workers: {self.worker_count}")
                self.logger.info(f"    Target directory: {self.working_dir}")

            for x in range(self.worker_count):
                if self.use_threading:
                    t = threading.Thread(target=restore_worker,
                                         args=(x, self.instance_id, self.working_dir))
                    t.setDaemon(True)
                    t.start()
                else:
                    p = Process(target=restore_worker, args=(x, self.instance_id, self.working_dir))
                    p.start()
                    self.plist.append(p)

            # Start done thread
            dt = threading.Thread(target=self.done_thread, args=('Restore',), name="Done thread")
            dt.setDaemon(True)
            dt.start()

            # Wait for workers to finish
            dt.join()
        except Exception as e:
            if self.logger:
                self.logger.exception(e)


def _string_getter(data) -> list[str]:
    if data is not None:
        return [data]
    else:
        return []


def _result_getter(data) -> list[str]:
    if data is not None:
        return [x for x in data if not x.endswith('.e')]
    else:
        return []


def _emptyresult_getter(data) -> list[str]:
    if data is not None:
        return [x for x in data if x.endswith('.e')]
    else:
        return []


def _error_getter(data) -> list[str]:
    if data is not None:
        return [x for x in data if x.rsplit('.e', 1)[1] not in ERROR_TYPES.values()]
    else:
        return []


def _sha256_getter(data) -> list[str]:
    if data is not None:
        return [x[:64] for x in data]
    else:
        return []


def _file_getter(data) -> list[str]:
    if data is not None:
        return [x['sha256'] for x in data]
    else:
        return []


def _result_file_getter(data) -> list[str]:
    if data is not None:
        supp = data.get("supplementary", []) + data.get("extracted", [])
        return _file_getter(supp)
    else:
        return []


FOLLOW_KEYS = {
    "alert": [
        ('submission', 'sid', _string_getter),
    ],
    "submission": [
        ('result', 'results', _result_getter),
        ('error', 'errors', _error_getter),
        ('file', 'results', _sha256_getter),
        ('file', 'files', _file_getter),
        ('file', 'errors', _sha256_getter),
    ],
    "results": [
        ('file', 'response', _result_file_getter),
    ]
}
